﻿using System;

namespace MicroLibrary
{
    /// <summary>
    /// MicroStopwatch class
    /// </summary>
    public class MicroStopwatch : System.Diagnostics.Stopwatch
    {
        readonly double _microSecPerTick = 1000000D / Frequency;

        public MicroStopwatch()
        {
            if (!System.Diagnostics.Stopwatch.IsHighResolution)
            {
                throw new Exception("On this system the high-resolution " +
                                    "performance counter is not available");
            }
        }

        public long ElapsedMicroseconds
        {
            get
            {
                return (long)(ElapsedTicks * _microSecPerTick);
            }
        }
    }

    /// <summary>
    /// MicroTimer class
    /// </summary>
    public class MicroTimer
    {
        public delegate void MicroTimerElapsedEventHandler(
                             object sender,
                             MicroTimerEventArgs timerEventArgs);
        public event MicroTimerElapsedEventHandler MicroTimerElapsed;

        System.Threading.Thread _threadTimer = null;
        long _ignoreEventIfLateBy = long.MaxValue;
        long _timerIntervalInMicroSec = 0;
        bool _stopTimer = true;

        public MicroTimer()
        {
        }

        public MicroTimer(long timerIntervalInMicroseconds)
        {
            Interval = timerIntervalInMicroseconds;
        }

        public long Interval
        {
            get { return _timerIntervalInMicroSec; }
            set { _timerIntervalInMicroSec = value; }
        }

        public long IgnoreEventIfLateBy
        {
            get
            {
                return _ignoreEventIfLateBy;
            }
            set
            {
                if (value <= 0)
                    _ignoreEventIfLateBy = long.MaxValue;
                else
                    _ignoreEventIfLateBy = value;
            }
        }

        public bool Enabled
        {
            set
            {
                if (value)
                    Start();
                else
                    Stop();
            }
            get
            {
                return (_threadTimer != null && _threadTimer.IsAlive);
            }
        }

        public void Start()
        {
            if (Enabled || Interval <= 0)
            {
                return;
            }

            _stopTimer = false;
            System.Threading.ThreadStart threadStart = delegate()
            {
                NotificationTimer(Interval, IgnoreEventIfLateBy, ref _stopTimer);
            };
            _threadTimer = new System.Threading.Thread(threadStart);
            _threadTimer.Priority = System.Threading.ThreadPriority.Highest;
            _threadTimer.Start();
        }

        public void Stop()
        {
            _stopTimer = true;

            if (_threadTimer.ManagedThreadId ==
                System.Threading.Thread.CurrentThread.ManagedThreadId)
            {
                return;
            }

            while (Enabled)
            {
                System.Threading.Thread.SpinWait(10);
            }
        }

        void NotificationTimer(long timerInterval,
                               long ignoreEventIfLateBy,
                               ref bool stopTimer)
        {
            int  timerCount = 0;
            long nextNotification = 0;

            MicroStopwatch microStopwatch = new MicroStopwatch();
            microStopwatch.Start();

            while (!stopTimer)
            {
                long callbackFunctionExecutionTime =
                    microStopwatch.ElapsedMicroseconds - nextNotification;
                nextNotification += timerInterval;
                timerCount++;
                long elapsedMicroseconds = 0;

                while ( (elapsedMicroseconds = microStopwatch.ElapsedMicroseconds)
                        < nextNotification)
                {
                    System.Threading.Thread.SpinWait(10);
                }

                long timerLateBy = elapsedMicroseconds - (timerCount * timerInterval);

                if (timerLateBy >= ignoreEventIfLateBy)
                {
                    continue;
                }

                MicroTimerEventArgs microTimerEventArgs =
                     new MicroTimerEventArgs(timerCount,
                                             elapsedMicroseconds,
                                             timerLateBy,
                                             callbackFunctionExecutionTime);
                MicroTimerElapsed(this, microTimerEventArgs);
            }

            microStopwatch.Stop();
        }
    }

    /// <summary>
    /// MicroTimer Event Argument class
    /// </summary>
    public class MicroTimerEventArgs : EventArgs
    {
        // Simple counter, number times timed event (callback function) executed
        private int _TimerCount;
        public int  TimerCount 
        {
            get { return _TimerCount; }
            private set{}
        }

        // Time when timed event was called since timer started
        private long _ElapsedMicroseconds;
        public long ElapsedMicroseconds 
        {
            get { return _ElapsedMicroseconds; }
            private set { }
        }

        // How late the timer was compared to when it should have been called
        private long _TimerLateBy;
        public long TimerLateBy 
        {
            get { return _TimerLateBy; }
            private set { }
        }

        // Time it took to execute previous call to callback function (OnTimedEvent)
        private long _CallbackFunctionExecutionTime;
        public long CallbackFunctionExecutionTime 
        {
            get { return _CallbackFunctionExecutionTime; }
            private set { }
        }

        public MicroTimerEventArgs(int  timerCount,
                                   long elapsedMicroseconds,
                                   long timerLateBy,
                                   long callbackFunctionExecutionTime)
        {
            _TimerCount = timerCount;
            _ElapsedMicroseconds = elapsedMicroseconds;
            _TimerLateBy = timerLateBy;
            _CallbackFunctionExecutionTime = callbackFunctionExecutionTime;
        }
    }
}

